<?php
/************************************************************************
 * Mailer v0.2.1-FINAL                                Start: 03/08/2011 *
 * ===================                          Last change: 03/08/2011 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : http-functions.php                               *
 * -------------------------------------------------------------------- *
 * Short description : HTTP-related functions                           *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : HTTP-relevante Funktionen                        *
 * -------------------------------------------------------------------- *
 * $Revision::                                                        $ *
 * $Date::                                                            $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author::                                                          $ *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * Copyright (c) 2009 - 2012 by Mailer Developer Team                   *
 * For more information visit: http://mxchange.org                      *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} // END - if

// Initialize HTTP handling
function initHttp () {
	// Initialize array
	$GLOBALS['http_header'] = array();
}

// Sends out all headers required for HTTP/1.1 reply
function sendHttpHeaders () {
	// Used later
	$now = gmdate('D, d M Y H:i:s') . ' GMT';

	// Send HTTP header
	addHttpHeader('HTTP/1.1 ' . getHttpStatus());

	// General headers for no caching
	addHttpHeader('Expires: ' . $now); // RFC2616 - Section 14.21
	addHttpHeader('Last-Modified: ' . $now);
	addHttpHeader('Cache-Control: no-store, no-cache, must-revalidate, pre-check=0, post-check=0, max-age=0'); // HTTP/1.1
	addHttpHeader('Pragma: no-cache'); // HTTP/1.0
	addHttpHeader('Connection: Close');

	// There shall be no output mode in raw output-mode
	if (!isRawOutputMode()) {
		// Send content-type not in raw output-mode
		addHttpHeader('Content-Type: ' . getContentType() . '; charset=UTF-8');
	} // END - if

	// Add language
	addHttpHeader('Content-Language: ' . getLanguage());
}

// Checks whether the URL is full-qualified (http[s]:// + hostname [+ request data])
function isFullQualifiedUrl ($url) {
	// Is there cache?
	if (!isset($GLOBALS[__FUNCTION__][$url])) {
		// Determine it
		$GLOBALS[__FUNCTION__][$url] = ((substr($url, 0, 7) == 'http://') || (substr($url, 0, 8) == 'https://'));
	} // END - if

	// Return cache
	return $GLOBALS[__FUNCTION__][$url];
}

// Generates the full GET URL from given base URL and data array
function generateGetUrlFromBaseUrlData ($baseUrl, $requestData = array()) {
	// Init URL
	$getUrl = $baseUrl;

	// Is it full-qualified?
	if (!isFullQualifiedUrl($getUrl)) {
		// Need to prepend a slash?
		if (substr($getUrl, 0, 1) != '/') {
			// Prepend it
			$getUrl = '/' . $getUrl;
		} // END - if

		// Prepend http://hostname from mxchange.org server
		$getUrl = getServerUrl() . $getUrl;
	} // END - if

	// Add data
	$body = http_build_query($requestData, '', '&');

	// There should be data, else we don't need to extend $baseUrl with $body
	if (!empty($body)) {
		// Is there a question-mark in the script?
		if (!isInString('?', $baseUrl)) {
			// No, so first char must be question mark
			$body = '?' . $body;
		} else {
			// Ok, add &
			$body = '&' . $body;
		}

		// Add script data
		$getUrl .= $body;

		// Remove trailed & to make it more conform
		if (substr($getUrl, -1, 1) == '&') {
			$getUrl = substr($getUrl, 0, -1);
		} // END - if
	} // END - if

	// Return it
	return $getUrl;
}

// Removes http[s]://<hostname> from given url
function removeHttpHostNameFromUrl ($url) {
	// Remove http[s]://
	$remove = explode(':', $url);
	$remove = explode('/', substr($remove[1], 3));

	// Remove the first element (should be the hostname)
	unset($remove[0]);

	// implode() back all other elements and prepend a slash
	$url = '/' . implode('/', $remove);

	// Return prepared URL
	return $url;
}

// Sends a HTTP request (GET, POST, HEAD are currently supported)
function sendHttpRequest ($requestType, $baseUrl, $requestData = array(), $removeHeader = FALSE) {
	// Init response
	$response = array();

	// Start "detecting" the request type
	switch ($requestType) {
		case 'HEAD': // Send a HTTP/1.1 HEAD request
			$response = sendHeadRequest($baseUrl, $requestData);
			break;

		case 'GET': // Send a HTTP/1.1 GET request
			$response = sendGetRequest($baseUrl, $requestData, $removeHeader);
			break;

		case 'POST': // Send a HTTP/1.1 POST request
			$response = sendPostRequest($baseUrl, $requestData, $removeHeader);
			break;

		default: // Unsupported HTTP request, this is really bad and needs fixing
			reportBug(__FUNCTION__, __LINE__, 'Unsupported request detected. requestType=' . $requestType . ',baseUrl=' . $baseUrl . ',requestData()=' . count($requestData));
			break;
	} // END - switch

	// Return response
	return $response;
}

// Sends a HEAD request
function sendHeadRequest ($baseUrl, $requestData = array()) {
	// Generate full GET URL
	$getUrl = generateGetUrlFromBaseUrlData($baseUrl, $requestData);

	// Is there http[s]:// in front of the URL?
	if (isFullQualifiedUrl($getUrl)) {
		// Remove http[s]://<hostname> from URL
		$getUrl = removeHttpHostNameFromUrl($getUrl);
	} elseif (substr($getUrl, 0, 1) != '/') {
		// Prepend a slash
		$getUrl = '/' . $getUrl;
	}

	// Extract hostname and port from script
	$host = extractHostnameFromUrl($baseUrl);

	// Generate HEAD request header
	$request  = 'HEAD ' . (isProxyUsed() === TRUE ? $baseUrl : '') . trim($getUrl) . ' HTTP/1.1' . getConfig('HTTP_EOL');
	$request .= 'Host: ' . $host . getConfig('HTTP_EOL');
	$request .= 'Referer: ' . getUrl() . '/admin.php' . getConfig('HTTP_EOL');
	if (isConfigEntrySet('FULL_VERSION')) {
		$request .= 'User-Agent: ' . getTitle() . '/' . getFullVersion() . getConfig('HTTP_EOL');
	} else {
		$request .= 'User-Agent: ' . getTitle() . '/' . getConfig('VERSION') . getConfig('HTTP_EOL');
	}
	$request .= 'Accept: image/png,image/*;q=0.8,text/plain,text/html,*/*;q=0.5' . getConfig('HTTP_EOL');
	$request .= 'Accept-Charset: UTF-8,*' . getConfig('HTTP_EOL');
	$request .= 'Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0' . getConfig('HTTP_EOL');
	$request .= 'Connection: close' . getConfig('HTTP_EOL');
	$request .= getConfig('HTTP_EOL');

	// Send the raw request
	$response = sendRawRequest($host, $request);

	// Return the result to the caller function
	return $response;
}

// Send a GET request
function sendGetRequest ($baseUrl, $requestData = array(), $removeHeader = FALSE) {
	// Generate full GET URL
	$getUrl = generateGetUrlFromBaseUrlData($baseUrl, $requestData);

	// Is there http[s]:// in front of the URL?
	if (isFullQualifiedUrl($getUrl)) {
		// Remove http[s]://<hostname> from url
		$getUrl = removeHttpHostNameFromUrl($getUrl);
	} elseif (substr($getUrl, 0, 1) != '/') {
		// Prepend a slash
		$getUrl = '/' . $getUrl;
	}

	// Extract hostname and port from script
	$host = extractHostnameFromUrl($baseUrl);

	// Generate GET request header
	$request  = 'GET ' . (isProxyUsed() === TRUE ? $baseUrl : '') . trim($getUrl) . ' HTTP/1.1' . getConfig('HTTP_EOL');
	$request .= 'Host: ' . $host . getConfig('HTTP_EOL');
	$request .= 'Referer: ' . getUrl() . '/admin.php' . getConfig('HTTP_EOL');
	if (isConfigEntrySet('FULL_VERSION')) {
		$request .= 'User-Agent: ' . getTitle() . '/' . getFullVersion() . getConfig('HTTP_EOL');
	} else {
		$request .= 'User-Agent: ' . getTitle() . '/' . getConfig('VERSION') . getConfig('HTTP_EOL');
	}
	$request .= 'Accept: image/png,image/*;q=0.8,text/plain,text/html,*/*;q=0.5' . getConfig('HTTP_EOL');
	$request .= 'Accept-Charset: UTF-8,*' . getConfig('HTTP_EOL');
	$request .= 'Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0' . getConfig('HTTP_EOL');
	$request .= 'Connection: close' . getConfig('HTTP_EOL');
	$request .= getConfig('HTTP_EOL');

	// Send the raw request
	$response = sendRawRequest($host, $request);

	// Should we remove header lines?
	if ($removeHeader === TRUE) {
		// Okay, remove them
		$response = removeHttpHeaderFromResponse($response);
	} // END - if

	// Return the result to the caller function
	return $response;
}

// Send a POST request, sometimes even POST requests have no parameters
function sendPostRequest ($baseUrl, $requestData = array(), $removeHeader = FALSE) {
	// Copy baseUrl to getUrl
	$getUrl = $baseUrl;

	// Is there http[s]:// in front of the URL?
	if (isFullQualifiedUrl($getUrl)) {
		// Remove http[s]://<hostname> from url
		$getUrl = removeHttpHostNameFromUrl($getUrl);
	} elseif (substr($getUrl, 0, 1) != '/') {
		// Prepend a slash
		$getUrl = '/' . $getUrl;
	}

	// Extract host name from script
	$host = extractHostnameFromUrl($baseUrl);

	// Construct request body
	$body = http_build_query($requestData, '', '&');

	// Generate POST request header
	$request  = 'POST ' . (isProxyUsed() === TRUE ? $baseUrl : '') . trim($baseUrl) . ' HTTP/1.0' . getConfig('HTTP_EOL');
	$request .= 'Host: ' . $host . getConfig('HTTP_EOL');
	$request .= 'Referer: ' . getUrl() . '/admin.php' . getConfig('HTTP_EOL');
	if (isConfigEntrySet('FULL_VERSION')) {
		$request .= 'User-Agent: ' . getTitle() . '/' . getFullVersion() . getConfig('HTTP_EOL');
	} else {
		$request .= 'User-Agent: ' . getTitle() . '/' . getConfig('VERSION') . getConfig('HTTP_EOL');
	}
	$request .= 'Accept: text/plain;q=0.8' . getConfig('HTTP_EOL');
	$request .= 'Accept-Charset: UTF-8,*' . getConfig('HTTP_EOL');
	$request .= 'Cache-Control: no-store, no-cache, must-revalidate, post-check=0, pre-check=0' . getConfig('HTTP_EOL');
	$request .= 'Content-Type: application/x-www-form-urlencoded' . getConfig('HTTP_EOL');
	$request .= 'Content-Length: ' . strlen($body) . getConfig('HTTP_EOL');
	$request .= 'Connection: close' . getConfig('HTTP_EOL');
	$request .= getConfig('HTTP_EOL');

	// Add body
	$request .= $body;

	// Send the raw request
	$response = sendRawRequest($host, $request);

	// Should we remove header lines?
	if ($removeHeader === TRUE) {
		// Okay, remove them
		$response = removeHttpHeaderFromResponse($response);
	} // END - if

	// Return the result to the caller function
	return $response;
}

// Sends a raw request (string) to given host (hostnames will be solved)
function sendRawRequest ($host, $request) {
	//* DEBUG: */ die('host='.$host.',request=<pre>'.$request.'</pre>');
	// Init errno and errdesc with 'all fine' values
	$errno = '0';
	$errdesc = '';

	// Default port is 80
	$port = 80;

	// Initialize array
	$response = array('', '', '');

	// Default is non-broken HTTP server implementation
	$GLOBALS['is_http_server_broken'] = FALSE;

	// Load include
	loadIncludeOnce('inc/classes/resolver.class.php');

	// Extract port part from host
	$portArray = explode(':', $host);
	if (count($portArray) == 2) {
		// Extract host and port
		$host = $portArray[0];
		$port = $portArray[1];
	} elseif (count($portArray) > 2) {
		// This should not happen!
		reportBug(__FUNCTION__, __LINE__, 'Invalid ' . $host . '. Please report this to the Mailer-Project team.');
	}

	// Get resolver instance
	$resolver = new HostnameResolver();

	// Get proxy host
	$proxyHost = compileRawCode(getProxyHost());

	// Open connection
	if (isProxyUsed() === TRUE) {
		// Resolve hostname into IP address
		$ip = $resolver->resolveHostname($proxyHost);

		// Connect to host through proxy connection
		$resource = fsockopen($ip, bigintval(getProxyPort()), $errno, $errdesc, 30);
	} else {
		// Resolve hostname into IP address
		$ip = $resolver->resolveHostname($host);

		// Connect to host directly
		$resource = fsockopen($ip, $port, $errno, $errdesc, 30);
	}
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'ip=' . $ip . ',host=' . $host . ',isProxyUsed()=' . intval(isProxyUsed()));

	// Is there a link?
	if (!is_resource($resource)) {
		// Failed!
		logDebugMessage(__FUNCTION__, __LINE__, $errdesc . ' (' . $errno . ')');
		return $response;
	} elseif ((!stream_set_blocking($resource, 0)) || (!stream_set_timeout($resource, 1))) {
		// Cannot set non-blocking mode or timeout
		logDebugMessage(__FUNCTION__, __LINE__, socket_strerror(socket_last_error()));
		return $response;
	}

	// Shall proxy be used?
	if (isProxyUsed() === TRUE) {
		// Setup proxy tunnel
		$response = setupProxyTunnel($host, $proxyHost, $port, $resource);

		// If the response is invalid, abort
		if ((count($response) == 3) && (empty($response[0])) && (empty($response[1])) && (empty($response[2]))) {
			// Invalid response!
			logDebugMessage(__FUNCTION__, __LINE__, 'Proxy tunnel not working?');
			return $response;
		} // END - if
	} // END - if

	// Write request
	fwrite($resource, $request);

	// Start counting
	$start = microtime(TRUE);

	// Read response
	while (!feof($resource)) {
		// Get info from stream
		$info = stream_get_meta_data($resource);

		// Is it timed out? 15 seconds is a really patient...
		if (($info['timed_out'] == TRUE) || (microtime(TRUE) - $start) > 15) {
			// Timeout
			logDebugMessage(__FUNCTION__, __LINE__, 'Timed out to get data from host ' . $host);

			// Abort here
			break;
		} // END - if

		// Get line from stream
		$line = fgets($resource, 128);

		// Ignore empty lines because of non-blocking mode
		if (empty($line)) {
			// uslepp a little to avoid 100% CPU load
			usleep(10);

			// Skip this
			continue;
		} // END - if

		// Check for broken HTTP implementations
		if (substr(strtolower($line), 0, 7) == 'server:') {
			// Anomic (see http://anomic.de, http://yacy.net) is currently broken
			$GLOBALS['is_http_server_broken'] = (count(getArrayKeysFromSubStrArray(strtolower($line), array('anomichttpd'))) > 0);
		} // END - if

		// Add it to response
		//* DEBUG: */ print 'line='.$line.'<br />';
		array_push($response, $line);
	} // END - while

	// Close socket
	fclose($resource);

	// Time request if debug-mode is enabled
	if (isDebugModeEnabled()) {
		// Add debug message...
		logDebugMessage(__FUNCTION__, __LINE__, 'Request took ' . (microtime(TRUE) - $start) . ' seconds and returned ' . count($response) . ' line(s).');
	} // END - if

	// Skip first empty lines
	$resp = $response;
	foreach ($resp as $idx => $line) {
		// Trim space away
		$line = trim($line);

		// Is this line empty?
		if (empty($line)) {
			// Then remove it
			array_shift($response);
		} else {
			// Abort on first non-empty line
			break;
		}
	} // END - foreach

	//* DEBUG: */ debugOutput('<strong>Request:</strong><pre>'.print_r($request, TRUE).'</pre>');
	//* DEBUG: */ debugOutput('<strong>Response:</strong><pre>'.print_r($response, TRUE).'</pre>');

	// Proxy agent found or something went wrong?
	if (!isset($response[0])) {
		// No response, maybe timeout
		$response = array('', '', '');
		logDebugMessage(__FUNCTION__, __LINE__, 'Invalid empty response array, maybe timed out?');
	} elseif ((substr(strtolower($response[0]), 0, 11) == 'proxy-agent') && (isProxyUsed() === TRUE)) {
		// Proxy header detected, so remove two lines
		array_shift($response);
		array_shift($response);
	} // END - if

	// Was the request successfull?
	if ((!isInStringIgnoreCase('200 OK', $response[0])) || (empty($response[0]))) {
		// Not found / access forbidden
		logDebugMessage(__FUNCTION__, __LINE__, 'Unexpected status code ' . $response[0] . ' detected. "200 OK" was expected.');
		$response = array('', '', '');
	} else {
		// Check array for chuncked encoding
		$response = unchunkHttpResponse($response);
	} // END - if

	// Return response
	return $response;
}

// Sets up a proxy tunnel for given hostname and through resource
function setupProxyTunnel ($host, $proxyHost, $port, $resource) {
	// Initialize array
	$response = array('', '', '');

	// Generate CONNECT request header
	$proxyTunnel  = 'CONNECT ' . $host . ':' . $port . ' HTTP/1.0' . getConfig('HTTP_EOL');
	$proxyTunnel .= 'Host: ' . $host . getConfig('HTTP_EOL');
	$proxyTunnel .= 'Proxy-Connection: Keep-Alive' . getConfig('HTTP_EOL');

	// Use login data to proxy? (username at least!)
	if (getProxyUsername() != '') {
		// Add it as well
		$encodedAuth = base64_encode(compileRawCode(getProxyUsername()) . ':' . compileRawCode(getProxyPassword()));
		$proxyTunnel .= 'Proxy-Authorization: Basic ' . $encodedAuth . getConfig('HTTP_EOL');
	} // END - if

	// Add last new-line
	$proxyTunnel .= getConfig('HTTP_EOL');
	//* DEBUG: */ debugOutput('<strong>proxyTunnel=</strong><pre>' . $proxyTunnel.'</pre>');

	// Write request
	fwrite($resource, $proxyTunnel);

	// Got response?
	if (feof($resource)) {
		// No response received
		return $response;
	} // END - if

	// Read the first line
	$resp = trim(fgets($resource, 10240));
	$respArray = explode(' ', $resp);
	if (((strtolower($respArray[0]) !== 'http/1.0') && (strtolower($respArray[0]) !== 'http/1.1')) || ($respArray[1] != '200')) {
		// Invalid response!
		return $response;
	} // END - if

	// All fine!
	return $respArray;
}

// Check array for chuncked encoding
function unchunkHttpResponse ($response) {
	// Default is not chunked
	$isChunked = FALSE;

	// Check if we have chunks
	foreach ($response as $line) {
		// Make lower-case and trim it
		$line = trim($line);

		// Entry found?
		if ((isInStringIgnoreCase('transfer-encoding', $line)) && (isInStringIgnoreCase('chunked', $line))) {
			// Found!
			$isChunked = TRUE;
			break;
		} // END - if
	} // END - foreach

	// Is it chunked?
	if ($isChunked === TRUE) {
		// Good, we still have the HTTP headers in there, so we need to get rid
		// of them temporarly
		//* DEBUG: */ die('<pre>'.htmlentities(print_r(removeHttpHeaderFromResponse($response), TRUE)).'</pre>');
		$tempResponse = http_chunked_decode(implode('', removeHttpHeaderFromResponse($response)));

		// We got a string back from http_chunked_decode(), so we need to convert it back to an array
		//* DEBUG: */ die('tempResponse['.strlen($tempResponse).']=<pre>'.replaceReturnNewLine(htmlentities($tempResponse)).'</pre>');

		// Re-add the headers
		$response = merge_array($GLOBALS['http_headers'], stringToArray(chr(10), $tempResponse));
	} // END - if

	// Return the unchunked array
	return $response;
}

// Removes HTTP header lines from a response array (e.g. output from send<Get|Post>Request() )
function removeHttpHeaderFromResponse ($response) {
	// Save headers for later usage
	$GLOBALS['http_headers'] = array();

	// The first array element has to contain HTTP
	if ((isset($response[0])) && (substr(strtoupper($response[0]), 0, 5) == 'HTTP/')) {
		// Okay, we have headers, now remove them with a second array
		$response2 = $response;
		foreach ($response as $line) {
			// Remove line
			array_shift($response2);

			// Add full line to temporary global array
			array_push($GLOBALS['http_headers'], $line);

			// Trim it for testing
			$lineTest = trim($line);

			// Is this line empty?
			if (empty($lineTest)) {
				// Then stop here
				break;
			} // END - if
		} // END - foreach

		// Write back the array
		$response = $response2;
	} // END - if

	// Return the modified response array
	return $response;
}

// Returns the flag if a broken HTTP server implementation was detected
function isBrokenHttpServerImplentation () {
	// Determine it
	$isBroken = ((isset($GLOBALS['is_http_server_broken'])) && ($GLOBALS['is_http_server_broken'] === TRUE));

	// ... and return it
	return $isBroken;
}

// Extract host from script name
function extractHostnameFromUrl (&$script) {
	// Use default SERVER_URL by default... ;) So?
	$url = getServerUrl();

	// Is this URL valid?
	if (substr($script, 0, 7) == 'http://') {
		// Use the hostname from script URL as new hostname
		$url = substr($script, 7);
		$extract = explode('/', $url);
		$url = $extract[0];
		// Done extracting the URL :)
	} // END - if

	// Extract host name
	$host = str_replace('http://', '', $url);
	if (isInString('/', $host)) {
		$host = substr($host, 0, strpos($host, '/'));
	} // END - if

	// Generate relative URL
	//* DEBUG: */ debugOutput('SCRIPT=' . $script);
	if (substr(strtolower($script), 0, 7) == 'http://') {
		// But only if http:// is in front!
		$script = substr($script, (strlen($url) + 7));
	} elseif (substr(strtolower($script), 0, 8) == 'https://') {
		// Does this work?!
		$script = substr($script, (strlen($url) + 8));
	}

	//* DEBUG: */ debugOutput('SCRIPT=' . $script);
	if (substr($script, 0, 1) == '/') {
		$script = substr($script, 1);
	} // END - if

	// Return host name
	return $host;
}

// Adds a HTTP header to array
function addHttpHeader ($header) {
	// Send the header
	//* DEBUG: */ logDebugMessage(__FUNCTION__ . ': header=' . $header);
	array_push($GLOBALS['http_header'], trim($header));
}

// Flushes all HTTP headers
function flushHttpHeaders () {
	// Is the header already sent?
	if (headers_sent()) {
		// Then abort here
		reportBug(__FUNCTION__, __LINE__, 'Headers already sent!');
	} elseif ((!isset($GLOBALS['http_header'])) || (!is_array($GLOBALS['http_header']))) {
		// Not set or not an array
		reportBug(__FUNCTION__, __LINE__, 'Headers not set or not an array, isset()=' . isset($GLOBALS['http_header']) . ', please report this.');
	}

	// Flush all headers if found
	foreach ($GLOBALS['http_header'] as $header) {
		// Send a single header
		header($header);
	} // END - foreach

	// Mark them as flushed
	$GLOBALS['http_header'] = array();
}

//-----------------------------------------------------------------------------
// Automatically re-created functions, all taken from user comments on www.php.net
//-----------------------------------------------------------------------------

if (!function_exists('http_build_query')) {
	// Taken from documentation on www.php.net, credits to Marco K. (Germany) and some light mods by R.Haeder
	function http_build_query($requestData, $prefix = '', $sep = '', $key = '') {
		$ret = array();
		foreach ((array) $requestData as $k => $v) {
			if (is_int($k) && $prefix != null) {
				$k = urlencode($prefix . $k);
			} // END - if

			if ((!empty($key)) || ($key === 0)) {
				$k = $key . '[' . urlencode($k) . ']';
			} // END - if

			if (is_array($v) || is_object($v)) {
				array_push($ret, http_build_query($v, '', $sep, $k));
			} else {
				array_push($ret, $k . '=' . urlencode($v));
			}
		} // END - foreach

		if (empty($sep)) {
			$sep = ini_get('arg_separator.output');
		} // END - if

		return implode($sep, $ret);
	}
} // END - if

if (!function_exists('http_chunked_decode')) {
	/**
	 * dechunk an HTTP 'transfer-encoding: chunked' message.
	 *
	 * @param	$chunk		The encoded message
	 * @return	$dechunk	The decoded message. If $chunk wasn't encoded properly reportBug() is being called
	 * @author	Marques Johansson (initial author)
	 * @author	Roland Haeder (heavy modifications and simplification)
	 * @link	http://php.net/manual/en/function.http-chunked-decode.php#89786
	 */
	function http_chunked_decode ($chunk) {
		// Detect multi-byte encoding
		$mbPrefix = detectMultiBytePrefix($chunk);
		//* DEBUG: */ print 'mbPrefix=' . $mbPrefix . '<br />';

		// Init some variables
		$offset = 0;
		$len = call_user_func_array($mbPrefix . 'strlen', array(($chunk)));
		$dechunk = '';

		// Walk through all chunks
		while ($offset < $len) {
			// Where does the \r\n begin?
			$lineEndAt = call_user_func_array($mbPrefix . 'strpos', array($chunk, getConfig('HTTP_EOL'), $offset));

			/* DEBUG: *
			print 'lineEndAt[<em>'.__LINE__.'</em>]='.$lineEndAt.'<br />
offset[<em>'.__LINE__.'</em>]='.$offset.'<br />
len='.$len.'<br />
next[offset,10]=<pre>'.replaceReturnNewLine(htmlentities(call_user_func_array($mbPrefix . 'substr', array($chunk, $offset, 10)))).'</pre>';
			/* DEBUG: */

			// Get next hex-coded chunk length
			$chunkLenHex = call_user_func_array($mbPrefix . 'substr', array($chunk, $offset, ($lineEndAt - $offset)));

			/* DEBUG: *
			print 'chunkLenHex[<em>'.__LINE__.'</em>]='.replaceReturnNewLine(htmlentities($chunkLenHex)).'<br />
';
			/* DEBUG: */

			// Validation if it is hexadecimal
			if (!isHexadecimal($chunkLenHex)) {
				// Please help debugging this
				//* DEBUG: */ die('ABORT:chunkLenHex=<pre>'.replaceReturnNewLine(htmlentities($chunkLenHex)).'</pre>');
				reportBug(__FUNCTION__, __LINE__, 'Value ' . $chunkLenHex . ' is no valid hexa-decimal string.');

				// This won't be reached
				return $chunk;
			} // END - if

			// Position of next chunk is right after \r\n
			$offset   = $offset + strlen($chunkLenHex) + strlen(getConfig('HTTP_EOL'));
			$chunkLen = hexdec(rtrim($chunkLenHex, getConfig('HTTP_EOL')));

			/* DEBUG: *
			print 'chunkLen='.$chunkLen.'<br />
offset[<em>'.__LINE__.'</em>]='.$offset.'<br />';
			/* DEBUG: */

			// Moved out for debugging
			$next  = call_user_func_array($mbPrefix . 'substr', array($chunk, $offset, $chunkLen));
			//* DEBUG: */ print 'next=<pre>'.replaceReturnNewLine(htmlentities($next)).'</pre>';

			/*
			 * Hack for e.g. YaCy HTTPDaemon (Anomic Server), this HTTP server
			 * is currently (revision 7567 and maybe earlier) broken and does
			 * not include the \r\n characters when it sents a "chunked" HTTP
			 * message.
			 */
			$count = 0;
			if (isBrokenHttpServerImplentation()) {
				// Count occurrences of \r\n
				$count = call_user_func_array($mbPrefix . 'substr_count', array($next, getConfig('HTTP_EOL')));
			} // END - if

			/*
			 * Correct chunk length because some broken HTTP server
			 * implementation subtract occurrences of \r\n in their chunk
			 * lengths.
			 */
			$chunkLen = hexdec(rtrim($chunkLenHex, getConfig('HTTP_EOL'))) - ($count * strlen(getConfig('HTTP_EOL')));

			// Add next chunk to $dechunk
			$dechunk .= call_user_func_array($mbPrefix . 'substr', array($chunk, $offset, $chunkLen));

			/* DEBUG: *
			print('offset[<em>'.__LINE__.'</em>]='.$offset.'<br />
lineEndAt[<em>'.__LINE__.'</em>]='.$lineEndAt.'<br />
len='.$len.'<br />
count='.$count.'<br />
chunkLen='.$chunkLen.'<br />
chunkLenHex='.$chunkLenHex.'<br />
dechunk=<pre>'.replaceReturnNewLine(htmlentities($dechunk)).'</pre>
chunk=<pre>'.replaceReturnNewLine(htmlentities($chunk)).'</pre>');
			/* DEBUG: */

			// Is $offset + $chunkLen larger than or equal $len?
			if (($offset + $chunkLen) >= $len) {
				// Then stop processing here
				break;
			} // END - if

			// Calculate offset of next chunk
			$offset = call_user_func_array($mbPrefix . 'strpos', array($chunk, getConfig('HTTP_EOL'), $offset + $chunkLen)) + 2;

			/* DEBUG: *
			print('offset[<em>'.__LINE__.'</em>]='.$offset.'<br />
next[100]=<pre>'.replaceReturnNewLine(htmlentities(call_user_func_array($mbPrefix . 'substr', array($chunk, $offset, 100)))).'</pre>
---:---:---:---:---:---:---:---:---<br />
');
			/* DEBUG: */
		} // END - while

		// Return de-chunked string
		return $dechunk;
	}
} // END - if

// Getter for request method
function getHttpRequestMethod () {
	// Console is default
	$requestMethod = 'console';

	// Is it set?
	if (isset($_SERVER['REQUEST_METHOD'])) {
		// Get current request method
		$requestMethod = $_SERVER['REQUEST_METHOD'];
	} // END - if

	// Return it
	return $requestMethod;
}

// Checks if 'content_type' is set
function isContentTypeSet () {
	return isset($GLOBALS['content_type']);
}

// Setter for content type
function setContentType ($contentType) {
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'contentType=' . $contentType);
	$GLOBALS['content_type'] = (string) $contentType;
}

// Getter for content type
function getContentType () {
	// Is it there?
	if (!isContentTypeSet()) {
		// Please fix this
		reportBug(__FUNCTION__, __LINE__, 'content_type not set in GLOBALS array.');
	} // END - if

	// Return it
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'content_type=' . $GLOBALS['content_type']);
	return $GLOBALS['content_type'];
}

// Logs wrong SERVER_NAME attempts
function logWrongServerNameRedirect () {
	// Is ext-sql_patches at least version 0.9.2?
	if (isExtensionInstalledAndNewer('sql_patches', '0.9.2')) {
		// Is there an entry?
		if (countSumTotalData(detectServerName(), 'server_name_log', 'server_name_id', 'server_name', TRUE, sprintf(" AND `server_name_remote_addr`='%s' AND `server_name_ua`='%s' AND `server_name_referrer`='%s'", SQL_ESCAPE(detectRemoteAddr(TRUE)), SQL_ESCAPE(detectUserAgent(TRUE)), SQL_ESCAPE(detectReferer(TRUE)))) == 1) {
			// Update counter, as all are the same
			SQL_QUERY_ESC("UPDATE
	`{?_MYSQL_PREFIX?}_server_name_log`
SET
	`server_name_counter`=`server_name_counter`+1
WHERE
	`server_name`='%s' AND
	`server_name_remote_addr`='%s' AND
	`server_name_ua`='%s' AND
	`server_name_referrer`='%s'
LIMIT 1",
				array(
					detectServerName(),
					detectRemoteAddr(TRUE),
					detectUserAgent(TRUE),
					detectReferer(TRUE)
				), __FUNCTION__, __LINE__);
		} else {
			// Then log it away
			SQL_QUERY_ESC("INSERT INTO `{?_MYSQL_PREFIX?}_server_name_log` (`server_name`, `server_name_remote_addr`, `server_name_ua`, `server_name_referrer`) VALUES('%s','%s', '%s', '%s')",
				array(
					detectServerName(),
					detectRemoteAddr(TRUE),
					detectUserAgent(TRUE),
					detectReferer(TRUE)
				), __FUNCTION__, __LINE__);
		}
	} // END - if
}

// [EOF]
?>
