<?php
/************************************************************************
 * Mailer v0.2.1-FINAL                                Start: 10/23/2009 *
 * ===================                          Last change: 10/23/2009 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : sql-functions.php                                *
 * -------------------------------------------------------------------- *
 * Short description : SQL functions to handle queries                  *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : SQL-Funktionen fuer Queries                      *
 * -------------------------------------------------------------------- *
 * $Revision::                                                        $ *
 * $Date::                                                            $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author::                                                          $ *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * Copyright (c) 2009 - 2012 by Mailer Developer Team                   *
 * For more information visit: http://mxchange.org                      *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} // END - if

// Init SQLs array
function initSqls () {
	// Init generic array
	setSqlsArray(array('generic' => array()));
}

// Checks whether the sqls array is initialized
function isSqlsInitialized () {
	return ((isset($GLOBALS['sqls'])) && (is_array($GLOBALS['sqls'])));
}

// Setter for SQLs array
function setSqlsArray ($SQLs) {
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'count()='.count($SQLs));
	$GLOBALS['sqls'] = (array) $SQLs;
}

// Remover for SQLs array
function unsetSqls () {
	unset($GLOBALS['sqls']);
}

// Getter for SQLs array
function getSqls () {
	return $GLOBALS['sqls'];
}

// Add an SQL to the list
function addSql ($sql) {
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, sprintf("sql=%s, count=%d", $sql, countSqls()));
	array_push($GLOBALS['sqls']['generic'], $sql);
}

// Merge SQLs together
function mergeSqls ($SQLs, $type = '') {
	// Should we merge full array or partial?
	if (empty($type)) {
		// Merge full array (may kill entries)
		setSqlsArray(merge_array(getSqls(), $SQLs));
	} else {
		// Merge sub array, so get it
		$array = getSqls();

		// Is the sub array there?
		if (isset($array[$type])) {
			// Then get it and merge it with the new one
			$array[$type] = merge_array($array[$type], $SQLs);
		} else {
			// Use new array
			$array[$type] = $SQLs;
		}

		// Call again..
		mergeSqls($array);
	}
}

// Counter for SQLs array
function countSqls () {
	// Default is false
	$count = '0';

	// Is the array there?
	if (isSqlsInitialized()) {
		// Then count it
		$count = count($GLOBALS['sqls']);
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, sprintf("count=%d", $count));
	} // END - if

	// Return it
	return $count;
}

// Checks whether the SQLs array is filled
function isSqlsValid () {
	//* DEBUG: */ debugOutput(__FUNCTION__ . ':' . intval(isSqlsInitialized()) . '/' . countSqls() . '/' . getCurrentExtensionName());
	return (
		(
			isSqlsInitialized()
		) && (
			countSqls() > 0
		)
	);
}

// Generates an updating SQL query from given array
function getUpdateSqlFromArray ($array, $tableName, $whereColumn, $whereData, $excludedFields, $multiDimId = NULL) {
	// Begin SQL query
	$SQL = 'UPDATE `{?_MYSQL_PREFIX?}_' . $tableName . '` SET ';

	// Insert all data
	foreach ($array as $entry => $value) {
		// Skip login/id entry
		if (in_array($entry, $excludedFields)) {
			continue;
		} // END - if

		// Is there a non-string (e.g. number, NULL, SQL function or back-tick at the beginning?
		if (is_null($multiDimId)) {
			// Handle one-dimensional data
			if (is_null($value)) {
				// NULL detected
				$SQL .= '`' . $entry . '`=NULL,';
			} elseif ((substr($value, -2, 2) == '()') || (substr($value, 0, 1) == '`')) {
				// SQL function needs no ticks (')
				$SQL .= '`' . $entry . '`=' . SQL_ESCAPE($value) . ',';
			} elseif ('' . bigintval($value, TRUE, FALSE) . '' == '' . $value . '')  {
				// No need for ticks (')
				$SQL .= '`' . $entry . '`=' . $value . ',';
			} else {
				// Strings need ticks (') around them
				$SQL .= '`' . $entry . "`='" . SQL_ESCAPE($value) . "',";
			}
		} else {
			// Handle multi-dimensional data
			if (is_null($value[$multiDimId])) {
				// NULL detected
				$SQL .= '`' . $entry . '`=NULL,';
			} elseif ((substr($value[$multiDimId], -2, 2) == '()') || (substr($value[$multiDimId], 0, 1) == '`')) {
				// SQL function needs no ticks (')
				$SQL .= '`' . $entry . '`=' . SQL_ESCAPE($value[$multiDimId]) . ',';
			} elseif (('' . bigintval($value[$multiDimId], TRUE, FALSE) . '' == '' . $value[$multiDimId] . ''))  {
				// No need for ticks (')
				$SQL .= '`' . $entry . '`=' . $value[$multiDimId] . ',';
			} else {
				// Strings need ticks (') around them
				$SQL .= '`' . $entry . "`='" . SQL_ESCAPE($value[$multiDimId]) . "',";
			}
		}
	} // END - foreach

	// Remove last 2 chars and finish query
	$SQL = substr($SQL, 0, -1) . ' WHERE `' . $whereColumn . '`=' . $whereData . ' LIMIT 1';

	// Return SQL query
	return $SQL;
}

// "Getter" for an "INSERT INTO" SQL query
function getInsertSqlFromArray ($array, $tableName) {
	// Init SQL
	$SQL = 'INSERT INTO
`{?_MYSQL_PREFIX?}_' . $tableName . '`
(
`' . implode('`, `', array_keys(postRequestArray())) . '`
) VALUES (';

	// Walk through all entries
	foreach (postRequestArray() as $key => $value) {
		// Log debug message
		//* NOISY-DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'tableName=' . $tableName . ',key=' . $key . ',value=' . $value);

		// Add all entries
		if (is_null($value)) {
			// Add NULL
			$SQL .= 'NULL,';
		} elseif (substr($value, -2, 2) == '()') {
			// SQL function needs no ticks (')
			$SQL .= SQL_ESCAPE($value) . ',';
		} elseif ('' . bigintval($value, TRUE, FALSE) . '' == '' . $value . '') {
			// Number detected, no need for ticks (')
			$SQL .= bigintval($value) . ',';
		} elseif ('' . (float) $value . '' == '' . $value . '') {
			// Float number detected
			$SQL .= sprintf('%01.5f', $value);
		} else {
			// Everything else might be a string, so add ticks around it
			$SQL .= chr(39) . SQL_ESCAPE($value) . chr(39) . ',';
		}
	} // END - foreach

	// Finish SQL query
	$SQL = substr($SQL, 0, -1) . ')';

	// Return SQL query
	return $SQL;
}

// Initializes the SQL link by bringing it up if set
function initSqlLink () {
	// Do this only if link is down
	assert(!SQL_IS_LINK_UP());

	// Is the configuration data set?
	if ((!empty($GLOBALS['mysql']['host'])) && (!empty($GLOBALS['mysql']['login'])) && (!empty($GLOBALS['mysql']['dbase']))) {
		// Remove cache
		unset($GLOBALS['is_sql_link_up']);

		// Connect to DB
		SQL_CONNECT($GLOBALS['mysql']['host'], $GLOBALS['mysql']['login'], $GLOBALS['mysql']['password'], __FILE__, __LINE__);

		// Is the link valid?
		if (SQL_IS_LINK_UP()) {
			// Enable exit on error
			enableExitOnError();

			// Is it a valid resource?
			if (SQL_SELECT_DB($GLOBALS['mysql']['dbase'], __FILE__, __LINE__) === TRUE) {
				// Set database name (required for ext-optimize and ifSqlTableExists())
				setConfigEntry('__DB_NAME', $GLOBALS['mysql']['dbase']);

				// Remove MySQL array from namespace
				unset($GLOBALS['mysql']);

				// Load cache
				loadIncludeOnce('inc/load_cache.php');
			} else {
				// Wrong database?
				reportBug(__FILE__, __LINE__, 'Wrong database selected.');
			}
		} else {
			// No link to database!
			reportBug(__FILE__, __LINE__, 'Database link is not yet up.');
		}
	} else {
		// Maybe you forgot to enter your database login?
		reportBug(__FILE__, __LINE__, 'Database login is missing.');
	}
}

// [EOF]
?>
